import re
from enum import Enum
from re import Match, Pattern
from typing import NamedTuple

separator = r"[\.,\s]*"

phil_regex = rf"""(?P<PHIL_PUB>
    phil
    {separator}
    (rep)?
    {separator} # 4 Phil. Rep., 545
)"""
scra_regex = r"(?P<SCRA_PUB>SCRA)"
offg_regex = rf"""(?P<OG_PUB>
                (
                    (
                        o
                        {separator}
                        g
                        {separator}
                    )|
                    (
                        off
                        {separator}
                        gaz(ette)?
                        {separator}
                    )
                )
                (
                    (
                        suppl? # Supp. vs. Suppl.; 47 Off. Gaz. Suppl. 12
                        {separator}
                    )|
                    (
                        \(? # 56 OG (No. 4) 1068
                        no # 49 O.G. No. 7, 2740 (1953),
                        {separator} # 46 O.G. No. 11, 90
                        \d{{1,4}}  # note enclosing brackets
                        \)?
                        {separator}
                    )
                )?
            )
        """


class PublisherStyle(NamedTuple):
    label: str
    description: str
    group_name: str
    regex_exp: str


class Publisher(Enum):
    PHIL = PublisherStyle(
        label="Phil.",
        description="Philippine Reports",
        group_name="PHIL_PUB",
        regex_exp=phil_regex,
    )
    SCRA = PublisherStyle(
        label="SCRA",
        description="Supreme Court Reports Annotated",
        group_name="SCRA_PUB",
        regex_exp=scra_regex,
    )
    OFFG = PublisherStyle(
        label="O.G.",
        description="Official Gazette",
        group_name="OG_PUB",
        regex_exp=offg_regex,
    )

    @property
    def pattern(self) -> Pattern:
        return re.compile(self.value.regex_exp, re.I | re.X)

    @classmethod
    def regex_options(cls) -> str:
        return rf"""
            (?P<publisher>
                {cls.SCRA.value.regex_exp}| # contains SCRA_PUB group name
                {cls.PHIL.value.regex_exp}| # contains PHIL_PUB group name
                {cls.OFFG.value.regex_exp} # contains OG_PUB group name
            )
        """

    @classmethod
    def all_patterns(cls) -> Pattern:
        return re.compile(cls.regex_options(), re.X | re.I)

    @classmethod
    def get_label(cls, match: Match):
        for _, m in cls.__members__.items():
            if match.group(m.value.group_name):
                return m.value.label



import re
from re import Pattern

from citation_date import report_date

from .publisher import Publisher

volume = r"""
    \b
    (?P<volume>
        [12]? # makes possible from 1000 to 2999
        \d{1,3}
        (
            \-A| # See Central Bank v. CA, 159-A Phil. 21, 34 (1975);
            a
        )?
    )
    \b
"""

page = r"""
    \b
    (?P<page>
        [12345]? # makes possible from 1000 to 5999
        \d{1,3}  # 49 Off. Gazette 4857
    )
    \b
"""

volpubpage = rf"""
    (?P<volpubpage>
        {volume}
        \s+
        {Publisher.regex_options()}
        \s+
        {page}
    )
"""

filler = r"""
    (?P<filler>
        [\d\-\.]{1,10}
    )
"""

extra = rf"""
    (?:
        (?:
            [\,\s,\-]*
            {filler}
        )?
        [\,\s]*
        {report_date}
    )?
"""

REPORT_REGEX = rf"{volpubpage}{extra}"

REPORT_PATTERN: Pattern = re.compile(REPORT_REGEX, re.X | re.I)



import datetime
from dataclasses import dataclass
from re import Match

from dateutil.parser import parse

from .publisher import Publisher
from .regex import REPORT_PATTERN


@dataclass
class Report:
    """Based on a `Match` object with pre-defined regex group names,
    Extract groups into respective fields.
    """

    match: Match
    publisher: str | None = None
    volpubpage: str | None = None
    volume: str | None = None
    page: str | None = None
    report_date: datetime.date | None = None

    def __post_init__(self):
        self.publisher = Publisher.get_label(self.match)
        self.volpubpage = self.match.group("volpubpage")
        self.volume = self.match.group("volume")
        self.page = self.match.group("page")
        self.report_date = self.convert_date_from_match(self.match)

    def __str__(self) -> str:
        return self.formal_report

    @property
    def formal_report(self):
        return f"{self.volume} {self.publisher} {self.page}"

    @property
    def phil(self):
        if self.publisher == Publisher.PHIL.value.label:
            return self.formal_report
        return None

    @property
    def scra(self):
        if self.publisher == Publisher.SCRA.value.label:
            return self.formal_report
        return None

    @property
    def offg(self):
        if self.publisher == Publisher.OFFG.value.label:
            return self.formal_report
        return None

    def convert_date_from_match(self, match: Match) -> datetime.date | None:
        if text := match.group("report_date"):
            try:
                return parse(text).date()
            except Exception:
                ...
        return None

    @classmethod
    def extract(cls, text: str):
        for match in REPORT_PATTERN.finditer(text):
            yield Report(match)

    @classmethod
    def extract_from_dict(cls, data: dict, report_type: str):
        """
        Assuming a dictionary with any of the following report_type keys
        `scra`, `phil` or `offg`, get the value of the Report property.
        >>> Report.extract_from_dict({"scra": "14 SCRA 314"}, "scra")
        '14 SCRA 314'
        """
        if report_type.lower() in ["scra", "phil", "offg"]:
            if candidate := data.get(report_type):
                try:
                    obj = next(cls.extract(candidate))
                    # will get the @property of the Report with the same name
                    if hasattr(obj, report_type):
                        return obj.__getattribute__(report_type)
                except StopIteration:
                    return None
        return None

    @classmethod
    def get_unique(cls, text: str) -> list[str]:
        """Will only get `Report` volpubpages (string) from the text"""
        reports = [r.volpubpage for r in cls.extract(text) if r.volpubpage]
        return list(set(reports))
